def compile_model(model, optimizer, loss=None, **kwargs):

    """
    Compiles a Keras model

    :param model: A keras model
    :param optimizer: The optimizer to use when compiling
    :param loss: The loss function to use when compiling
    :param kwargs: Additional arguments to pass to the compiler
    :return: a (compiled) Keras model
    """


    model.compile(optimizer=optimizer,
                  loss=loss,
                  **kwargs)

    return model



import math

def evaluate(model, data_set, dataset_name, metric_names=None):
    """
    Generates the train, valid or test performance metrics of a trained model

    :param model: an H2O model object
    :param data_set: A DataSet with H2ODataWrappers or H2OSparklingDataWrappers as "full_data" attribute
    :param dataset_name: "train", "valid" or "test", indicating the data being evaluated on
    :return: Dictionary of the form {"metric_name":"metric_value"}
    """

    h2o_df = data_set.full_data.underlying

    if not metric_names:
        metric_names = [
            "logloss", "mse", "mae", "mean_residual_deviance", "rmse", "rmsle", "residual_deviance",
            "residual_degrees_of_freedom", "null_deviance", "null_degrees_of_freedom", "aic", "auc", "gini", "r2"
        ]

    # http://docs.h2o.ai/h2o/latest-stable/h2o-py/docs/_modules/h2o/model/model_base.html
    if dataset_name == "train":
        performance = model.model_performance(test_data=None, train=True, valid=False)
    elif dataset_name == "valid":
        performance = model.model_performance(test_data=None, train=False, valid=True)
    elif dataset_name == "test":
        performance = model.model_performance(test_data=h2o_df, train=False, valid=False)
    else:
        raise ValueError(
            "Invalid dataset type. Expecting 'train', 'valid' or 'test'")

    metrics = {}
    for metric_name in metric_names:

        # get all metrics. Ones that have not been calculated are return as None
        metrics[metric_name] = _try_param(getattr(performance, metric_name))

        # remove metrics that have no present calculation
        if metrics[metric_name] is None:
            del metrics[metric_name]

    return metrics

def evaluate_threshold_metrics(model, data_set, dataset_name, threshold_metric_names=None):
    """
    Generates the train, valid or test performance metrics of a trained model

    :param model: an H2O model object
    :param data_set: A DataSet with H2ODataWrappers or H2OSparklingDataWrappers as "full_data" attribute
    :param dataset_name: "train", "valid" or "test", indicating the data being evaluated on
    :return: Dictionary of the form {"metric_name":"metric_value"}
    """

    if not threshold_metric_names:
        #TODO must filter on threshold_metric_names
        threshold_metric_names= [
            "f1", "f2", "f0point5", "accuracy", "precision", "recall", "specificity", "absolute_mcc",
            "min_per_class_accuracy","mean_per_class_accuracy", "tns", "fns", "fps", "tps", "tnr", "fnr",
            "fpr", "tpr"
        ]


    h2o_df = data_set.full_data.underlying

    if dataset_name == "train":
        performance = model.model_performance(
            test_data=None, train=True, valid=False)
    elif dataset_name == "valid":
        performance = model.model_performance(
            test_data=None, train=False, valid=True)
    elif dataset_name == "test":
        performance = model.model_performance(
            test_data=h2o_df, train=False, valid=False)
    else:
        raise ValueError(
            "Invalid dataset type. Expecting 'train', 'valid' or 'test'")

    if performance._metric_json.get("thresholds_and_metric_scores"):
        threshold_metrics = performance._metric_json["thresholds_and_metric_scores"].as_data_frame().to_dict()
    else:
        threshold_metrics = {}
    return threshold_metrics










def _try_param(get_param):
    """
    Tries to execute the function "get_param" and returns its value. If this fails it returns None

    :param get_param: A function such as "logloss", "mse" etc. that is used to fetch the standard H2O metics
    :return:
    """

    try:
        if math.isnan(get_param()):
            return -9999.0000
        else:
            return get_param()
    except Exception:
        return None


from keras.layers import Dense
from keras.models import Sequential

def define_model(dense_activation, final_activation, nb_classes, nb_features, seed=1234):
    """
    A small example function for defining a simple (mlp) model. This just serves to illustrate what a "define_model"
    function might look like. It is expected that users will create their own "define_model" functions to use together with
    the other mercury-ml functions

    :param string dense_activation: The activation function to use in the dense layer.
    :param string final_activation: The activation function to use in the final layer.
    :param int nb_classes: The number of classes.
    :param int nb_features: The number of features.
    :param int seed: The random seed.
    :return:
    """

    model = Sequential()
    model.add(Dense(24, activation=dense_activation, input_shape=(nb_features,)))
    model.add(Dense(nb_classes, activation=final_activation))

    return model



def predict(data_set, model, return_columns = None, **kwargs):
    """
    Predicts with a trained model and returns the results as an H2ODataWrapper or H2OSparklingDataWrapper

    :param DataSet data_set: A DataSet with H2ODataWrappers or H2OSparklingDataWrappers as "full_data" attribute
    :param model: An H2O Model
    :param list return_columns: A list of columns to return from the resulting prediction
    :param kwargs: Additional keyword arguments to be passed to model.predict
    :return: And H2ODataWrapper or H2OSparklingDataWrapper with predictions
    """

    h2o_df = data_set.full_data.underlying
    prediction_h2o_df = model.predict(h2o_df, **kwargs)

    if return_columns:
        prediction_h2o_df = prediction_h2o_df[return_columns]

    from mercury_ml.common.providers.data_wrappers.h2o import H2ODataWrapper
    #TODO this should return an H2OSparklingDataWrapper in the input type is H2OSparklingDataWrapper
    return H2ODataWrapper(underlying=prediction_h2o_df, field_names=list(prediction_h2o_df.names))


def get_keras_optimizer(optimizer_name, optimizer_params):
    """
    Fetches a vanilla Keras model optimizer

    :param string optimizer_name: The name of the optimizer
    :param dict optimizer_params: The parameters with which the optimizer should be initialized
    :return:
    """

    from keras import optimizers

    return getattr(optimizers, optimizer_name)(**optimizer_params)



class PandasDataWrapper():
    """
    A DataWrapper with a Pandas DataFrame as its underlying data
    """

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names

    def slice_on_column_names(self, column_names):
        """
        Returns a Pandas DataFrame consisting of the columns that are present in the column_names list

        :param list list column_names: the names of the columns to return
        :return: Pandas DataFrame consisting of the selected columns
        """

        return self.underlying[column_names]

    def to_h2o(self, factor_columns_list=None):
        """
        Creates a new H2ODataWrapper based on this PandasDataWrapper and returns it as a new object

        :params list factor_columns_list: The names of the columns that should be treated as factors
        :return: a new H2ODataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.h2o import H2ODataWrapper
        import h2o
        df_h2o = h2o.H2OFrame(self.underlying)

        # convert columns
        if not factor_columns_list:
            factor_columns_list = []

        for factor_column_name in factor_columns_list:
            df_h2o[factor_column_name] = df_h2o[factor_column_name].asfactor()

        return H2ODataWrapper(df_h2o, self.field_names)

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this PandasDataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
        return NumpyDataWrapper(self.underlying.values, self.field_names)

    def to_spark(self, spark_session_params=None):
        """
        Creates a new SparkDataWrapper based on this PandasDataWrapper and returns it as a new object

        :return: a new SparkDataWrapper
        :param: dict spark_session_params: A dictionary of parameters according to which to get or initialize a Spark session
        """

        from mercury_ml.common.providers.data_wrappers.spark import SparkDataWrapper
        from mercury_ml.spark.providers.session import get_or_create_spark_session
        self.underlying.columns = self.underlying.columns.astype(str)

        if not spark_session_params:
            spark_session_params = {}
        spark = get_or_create_spark_session(**spark_session_params)

        return SparkDataWrapper(
            underlying=spark.createDataFrame(self.underlying),
            field_names=self.field_names
        )

    def to_h2o_sparkling(self, spark_session_params=None, factor_columns_list=None, h2o_sparkling_params=None):
        """
        Creates a new H2ODataWrapper based on this PandasDataWrapper and returns it as a new object

        :param list factor_columns_list: The names of the columns that should be treated as factors
        :param list spark_session_params: A dictionary of parameters according to which to get or initialize a Spark session
        :param list h2o_sparkling_params: A dictionary of parameters according to which to get or initialize an H2OSparkling session

        :return: a new H2ODataWrapper
        """

        if spark_session_params is None:
            spark_session_params = {}

        if h2o_sparkling_params is None:
            h2o_sparkling_params = {}

        return self.to_spark(spark_session_params).to_h2o_sparkling(factor_columns_list=factor_columns_list,
                                                                    h2o_sparkling_params=h2o_sparkling_params)

    def to_pandas(self):
        return self

    def concatenate(self, right_data_wrapper):
        """
        Concatenates this PandasDataWrapper with the one given in right_data_wrapper and returns a new PandasDataWrapper

        :param PandasDataWrapper right_data_wrapper: to be concatenated to the right of "self"
        :return: a new PandasDataWrapper
        """

        import pandas as pd
        new_underlying = pd.concat([self.underlying, right_data_wrapper.underlying], axis=1)
        new_field_names = self.field_names + right_data_wrapper.field_names
        return PandasDataWrapper(
            underlying=new_underlying,
            field_names=new_field_names
        )




import json
import os

def save_keras_hdf5(model, filename, local_dir, extension=None):
    """
    Saves a Keras model in .h5 format

    :param model: A Keras model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The filepath where the model is saved
    """

    if not extension:
        extension = ".h5"

    _make_dirs(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir + "/" + filename)
    model.save(path)

    return path

def save_tensorflow_serving_predict_signature_def(model, local_dir, filename, input_name="input", output_name="output",
                                                  extension=".zip", model_number="1", do_save_labels_txt=False,
                                                  labels_list=None, temp_base_dir=None):
    """
    Saves a Kers model in the format expected by TensorFlow serving (put various objects in a folder structure and
    zips it)

    :param model: A Keras model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param input_name: The name of the input tensors
    :param output_name: The name of the output tensors
    :param string extension: Extension to the filename with which the model is to be saved
    :param string model_number: The model version number
    :param bool do_save_labels_txt: In true a txt file with the model labels will be included in with the saved model files
    :param list labels_list: The labels to be saved
    :param string temp_base_dir: Temporary local directory where the model objects are to be saved. They will be compressed and moved
    to local_dir thereafter
    :return: The filepath where the model is saved
    """

    from keras import backend as K
    from tensorflow.python.saved_model import builder as saved_model_builder
    from tensorflow.python.saved_model.signature_def_utils import predict_signature_def
    from tensorflow.python.saved_model import tag_constants

    if not temp_base_dir:
        temp_base_dir = os.path.join(os.getcwd(), "_tmp_model", filename)

    if not os.path.isdir(temp_base_dir):
        os.makedirs(temp_base_dir)

    # Set the learning phase to Test since the model is already trained.
    K.set_learning_phase(0)

    temp_dir = os.path.join(temp_base_dir, model_number)

    builder = saved_model_builder.SavedModelBuilder(temp_dir)

    # Create prediction signature to be used by TensorFlow Serving Predict API
    signature = predict_signature_def(inputs={input_name: model.input},
                                      outputs={output_name: model.output})

    sess = K.get_session()
    # with K.get_session() as sess:
    # Save the meta graph and the variables
    builder.add_meta_graph_and_variables(sess=sess, tags=[tag_constants.SERVING],
                                         signature_def_map={"predict": signature, "serving_default": signature})

    builder.save()

    if do_save_labels_txt:
        save_labels_txt(filename="labels", local_dir=temp_dir, labels_list=labels_list)

    # create zip file
    import shutil
    shutil.make_archive(os.path.join(local_dir, filename), extension[1:], temp_base_dir)

    # remove temporary folder
    shutil.rmtree(temp_base_dir, ignore_errors=True)

    return os.path.join(local_dir, filename + extension)



def save_tensorflow_graph(model, filename, local_dir, num_outputs = 1, quantize = False, extension=None):
    """
    Saves a Keras model as a TensorFlow graph

    :param model: A Keras model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param int num_outputs: The number of outputs in the model
    :param bool quantize: If set to true will try to quantize the graph (could improve performance)
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The filepath where the model is saved
    """

    if not extension:
        extension = ".pb"

    _make_dirs(local_dir)

    filename = filename + extension
    path = os.path.join(local_dir + "/" + filename)
    
    tf_graph = _get_tensorflow_graph(model, num_outputs, quantize)

    from tensorflow.python.framework import graph_io
    graph_io.write_graph(tf_graph, local_dir, filename, as_text=False)

    return path

def save_tensorrt_pbtxt_config(model, filename, local_dir, model_base_name, extension=None):
    """
    Derives a pbtxt config from a Keras model and saves it

    :param model: A Keras model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param model_base_name: The base name of the model, to be used in the pbtxt config
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The filepath where the model is saved
    """

    if not extension:
        extension = ".pbtxt"

    _make_dirs(local_dir)

    filename = filename + extension
    path = os.path.join(local_dir + "/" + filename)
    
    config = _get_tensorrt_pbtxt_config_storage_model_format(model, model_base_name)

    with open(path, "w") as f:
        f.write(config)

    return path

def save_tensorrt_json_config(model, filename, local_dir, extension=None):
    """
    Derives a json config from a Keras model and saves it

    :param model: A Keras model
    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The filepath where the model is saved
    """

    if not extension:
        extension = ".json"

    _make_dirs(local_dir)
    filename = filename + extension
    path = os.path.join(local_dir + "/" + filename)
    
    config = _get_tensorrt_json_config_storage_model_format(model, filename)
    with open(path, "w") as f:
        _ = json.dump(config, f, sort_keys=True, indent=2)

    return path


def save_labels_txt(filename, local_dir, labels_list, model=None, extension=None):
    """

    :param string local_dir: Local directory where the model is to be saved
    :param string filename: Filename with which the model is to be saved
    :param labels_list:
    :param model:
    :param string extension: Extension to the filename with which the model is to be saved
    :return: The filepath where the model is saved
    """

    if not extension:
        extension = ".txt"

    _make_dirs(local_dir)

    filename = filename + extension
    path = os.path.join(local_dir + "/" + filename)
    
    with open(path, "w") as f:
        for item in labels_list:
            f.write("%s\n" % item)

    return path



def _get_tensorflow_graph(model, num_outputs = 1, quantize = False):
    output_node_prefix = "output_node" #TODO check if this can be replaced with just using the Keras output name?
    import tensorflow as tf
    from keras import backend as K

    K.set_learning_phase(0)

    pred = [None] * num_outputs
    pred_node_names = [None] * num_outputs
    for i in range(num_outputs):
        pred_node_names[i] = output_node_prefix + str(i)
    pred[i] = tf.identity(model.outputs[i], name=pred_node_names[i])
    print('output nodes names are: ', pred_node_names)

    sess = K.get_session()

    from tensorflow.python.framework import graph_util

    if quantize:
        from tensorflow.tools.graph_transforms import TransformGraph
        transforms = ["quantize_weights", "quantize_nodes"]
        transformed_graph_def = TransformGraph(sess.graph.as_graph_def(), [], pred_node_names, transforms)
        constant_graph = graph_util.convert_variables_to_constants(sess, transformed_graph_def, pred_node_names)
    else:
        constant_graph = graph_util.convert_variables_to_constants(sess, sess.graph.as_graph_def(), pred_node_names)

    return constant_graph

def _get_tensorrt_pbtxt_config_storage_model_format(model, model_base_name):
    input_name = model.input.name.split(":")[0]
    output_name = model.output.name.split(":")[0]
    model_type = "tensorflow_graphdef"
    input_dims = str(list(model.layers[0].input_shape[1:]))
    output_dims = str(list(model.layers[-1].output_shape[1:]))
    config = \
        'name: "{}"\n'.format(model_base_name) + \
        'platform: "{}"\n'.format(model_type) + \
        'max_batch_size: 128\n' + \
        'input [\n' + \
        '   {\n' + \
        '       name: "{}"\n'.format(input_name) + \
        '       data_type: TYPE_FP32\n' + \
        '       model_format: model_format_NHWC\n' + \
        '       dims: {}\n'.format(input_dims) + \
        '   }\n' + \
        ']\n' + \
        'output [\n' + \
        '   {\n' + \
        '       name: "{}"\n'.format(output_name) + \
        '       data_type: TYPE_FP32\n' + \
        '       dims: {}\n'.format(output_dims) + \
        '       label_filename: "labels.txt"\n' + \
        '   }\n' + \
        ']\n'

    return config


def _get_tensorrt_json_config_storage_model_format(model, filename):
    input_name = model.input.name.split(":")[0]
    output_name = model.output.name.split(":")[0]
    model_type = "tensorflow_graphdef"
    input_dims = list(model.layers[0].input_shape[1:])
    output_dims = list(model.layers[-1].output_shape[1:])
    config = {
                "name": filename,
                "platform": model_type,
                "max_batch_size": 128,
                "input": [
                   {
                       "name": input_name,
                       "data_type": "TYPE_FP32",
                       "model_format": "model_format_NHWC",
                       "dims": input_dims
                   }
                ],
                "output": [
                   {
                       "name": output_name,
                       "data_type": "TYPE_FP32",
                       "dims": output_dims,
                       "label_filename": "labels.txt"
                   }
                ]
            }

    return config


def _make_dirs(dir):
    import os
    if not os.path.exists(dir):
        os.makedirs(dir)



import numpy as np

def get_random_eraser(p=0.5, s_l=0.02, s_h=0.4, r_1=0.3, r_2=1/0.3, v_l=0, v_h=255, pixel_level=False):
    """
    A preprocessor that performs random erasure ob pixels.
    """
    def random_eraser(input_img):
        img_h, img_w, img_c = input_img.shape
        p_1 = np.random.rand()

        if p_1 > p:
            return input_img

        while True:
            s = np.random.uniform(s_l, s_h) * img_h * img_w
            r = np.random.uniform(r_1, r_2)
            w = int(np.sqrt(s / r))
            h = int(np.sqrt(s * r))
            left = np.random.randint(0, img_w)
            top = np.random.randint(0, img_h)

            if left + w <= img_w and top + h <= img_h:
                break

        if pixel_level:
            c = np.random.uniform(v_l, v_h, (h, w, img_c))
        else:
            c = np.random.uniform(v_l, v_h)

        input_img[top:top + h, left:left + w, :] = c

        return input_img

    return random_eraser


import os
import json

def store_h2o_frame(data, directory, filename, force=False, parts=1):
    """
    Export a given H2OFrame to a path on the machine this python session is currently connected to.

    :param data: the Frame to save to disk.
    :param directory: the directory to the save point on disk.
    :param filename: the name to save the frame to.
    :param force: if True, overwrite any preexisting file with the same path
    :param parts: enables export to multiple 'part' files instead of just a single file.
        Convenient for large datasets that take too long to store in a single file.
        Use parts=-1 to instruct H2O to determine the optimal number of part files or
        specify your desired maximum number of part files. Path needs to be a directory
        when exporting to multiple files, also that directory must be empty.
        Default is ``parts = 1``, which is to export to a single file.
    :return string filepath: the path to which the file was stored.
    """

    if not os.path.isdir(directory):
        os.makedirs(directory)

    filepath = _make_local_path(os.path.join(directory, filename))

    from h2o.job import H2OJob
    from h2o.utils.typechecks import assert_is_type
    from h2o.frame import H2OFrame
    from h2o import api
    assert_is_type(data, H2OFrame)
    assert_is_type(filepath, str)
    assert_is_type(force, bool)
    assert_is_type(parts, int)
    H2OJob(api("POST /3/Frames/%s/export" % (data.frame_id), data={"path": filepath, "num_parts": parts, "force": force}),
           "Export File").poll()

    return filepath


def store_pandas_pickle(data, directory, filename, compression=None):
    """
    Saves data to Disk as Pickle file.

    :param DataFrame data: The data to be saved
    :param string dir: dir where the data file should be saved
    :param string filename: Name that the data file should be saved with
    :return string filepath: the path to which the file was stored.
    """

    if not compression:
        compression_postfix = ""
    else:
        compression_postfix = "." + compression

    if not os.path.exists(directory):
        os.makedirs(directory)

    filepath = os.path.join(directory, filename + ".pkl")

    data.to_pickle(path=os.path.join(directory, filename + ".pkl" + compression_postfix),
                   compression=compression)

    return filepath


def store_pandas_json(data, directory, filename, orient="table", compression=None):
    """
    Saves data to Disk as JSON file.

    :param DataFrame data: The data to be saved
    :param string dir: dir where the data file should be saved
    :param string filename: Name that the data file should be saved with
    :return string filepath: the path to which the file was stored.
    """

    if compression is None:
        compression_postfix = ""
    else:
        compression_postfix = "." + compression

    if not os.path.exists(directory):
        os.makedirs(directory)

    filepath = os.path.join(directory, filename + ".json")

    data.to_json(path_or_buf=filepath,
                 orient=orient)

    return filepath


def store_dict_json(data, directory, filename):
    """
    Saves data to Disk as JSON file.

    :param dict data: The data to be saved
    :param string dir: dir where the data file should be saved
    :param string filename: Name that the data file should be saved with
    :return string filepath: the path to which the file was stored.
    """

    if not os.path.isdir(directory):
        os.makedirs(directory)

    filepath = os.path.join(directory, filename)
    with open(filepath, "w") as f:
        json.dump(data, f, indent=2)

    return filepath


def _make_local_path(path_name):
    if path_name[0] == ".":
        path_name = os.path.join(os.getcwd(), path_name)
        path_name = os.path.abspath(path_name)
    return path_name



"""
Simple IoC containers that provide direct access to various Keras providers
"""


class ModelSavers:
    from mercury_ml.keras.providers import model_saving
    save_hdf5 = model_saving.save_keras_hdf5
    save_tensorflow_graph = model_saving.save_tensorflow_graph
    save_tensorrt_pbtxt_config = model_saving.save_tensorrt_pbtxt_config
    save_tensorrt_json_config = model_saving.save_tensorrt_json_config
    save_labels_txt = model_saving.save_labels_txt
    save_tensorflow_serving_predict_signature_def = model_saving.save_tensorflow_serving_predict_signature_def

class ModelLoaders:
    from mercury_ml.keras.providers import model_loading
    load_hdf5 = model_loading.load_hdf5_model

class LossFunctionFetchers:
    from mercury_ml.keras.providers import loss_function_fetching
    get_keras_loss = loss_function_fetching.get_keras_loss
    get_custom_loss = loss_function_fetching.get_custom_loss

class OptimizerFetchers:
    from mercury_ml.keras.providers import optimizer_fetching
    get_keras_optimizer = optimizer_fetching.get_keras_optimizer

class ModelCompilers:
    from mercury_ml.keras.providers import model_compilation
    compile_model = model_compilation.compile_model


class ModelFitters:
    from mercury_ml.keras.providers import model_fitting
    fit = model_fitting.fit
    fit_generator = model_fitting.fit_generator

class ModelDefinitions:
    from mercury_ml.keras.providers.model_definition import conv_simple, mlp_simple

    # these are just two small example model definitions. Users should define their own models
    # to use as follows:
    # >>> ModelDefinitions.my_model = my_model_module.define_model

    define_conv_simple = conv_simple.define_model
    define_mlp_simple = mlp_simple.define_model


class GeneratorPreprocessingFunctionGetters:
    from mercury_ml.keras.providers.generator_preprocessors import get_random_eraser
    get_random_eraser = get_random_eraser


class CallBacks:
    from mercury_ml.keras.providers.model_callbacks import  TensorBoardProvider, \
        BaseLoggerProvider, EarlyStoppingProvider, ModelCheckpointProvider, TerminateOnNaNProvider, \
        ProgbarLoggerProvider, RemoteMonitorProvider, LearningRateSchedulerProvider, ReduceLROnPlateauProvider, \
        CSVLoggerProvider

    tensorboard = TensorBoardProvider
    base_logger = BaseLoggerProvider
    terminate_on_nan = TerminateOnNaNProvider
    progbar_logger = ProgbarLoggerProvider
    model_checkpoint = ModelCheckpointProvider
    early_stopping = EarlyStoppingProvider
    remote_monitor = RemoteMonitorProvider
    learning_rate_scheduler = LearningRateSchedulerProvider
    reduce_lr_on_plateau = ReduceLROnPlateauProvider
    csv_logger = CSVLoggerProvider

class ModelEvaluators:
    from mercury_ml.keras.providers import model_evaluation
    evaluate = model_evaluation.evaluate
    evaluate_generator = model_evaluation.evaluate_generator

class PredictionFunctions:
    from mercury_ml.keras.providers import prediction
    predict = prediction.predict
    predict_generator = prediction.predict_generator



"""Classes that facilitate moving data from one store into another"""

def singleton(TheClass):
    """ decorator for a class to make a singleton out of it """
    class_instances = {}

    def get_instance(*args, **kwargs):
        """ creating or just return the one and only class instance.
            The singleton depends on the parameters used in __init__ """
        key = (TheClass, args, str(kwargs))
        if key not in class_instances:
            class_instances[key] = TheClass(*args, **kwargs)
        return class_instances[key]

    return get_instance

@singleton
class S3Singleton:
    def __init__(self, **kwargs):
        import boto3
        session = boto3.Session(**kwargs)
        self.s3 = session.resource("s3")



import os
import subprocess

def copy_from_hdfs_to_disk(source_dir, target_dir, filename, overwrite=False, delete_source=False):
    """
    Moves a file from HDFS to Disk

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :return:
    """

    # TODO change this to use the Python "hdfs" class
    subprocess.call(["hdfs", "dfs", "-copyToLocal", os.path.join(source_dir,
                                                                 filename), os.path.join(target_dir, filename)])



import os

def copy_from_s3_to_disk(source_dir, target_dir, filename, overwrite=False, delete_source=False, s3_session_params=None,
                         reuse_existing=True):
    """
    Moves a file from S3 to Disk

    :param string source_dir: Path of file to be copied
    :param string target_dir: Path that file is to be copied to
    :param string filename: Name of file to be copied
    :param bool overwrite: If true, overwrite if file already exists in target_dir
    :param bool delete_source: If true, delete file from source_dir after copying
    :param dict s3_session_params: a dictionary of parameters used to initialise a new S3 session
    :param bool reuse_existing_s3_session: If set to True an existing S3 session will be reused instead of opening a new one
    :return:
    """

    if not s3_session_params:
        s3_session_params={}

    if not overwrite and os.path.isfile(target_dir + "/" + filename):
        pass
    else:
        import boto3
        if not reuse_existing:
            session = boto3.Session(**s3_session_params)
            s3 = session.resource("s3")
        else:
            from mercury_ml.common.providers.artifact_copying import S3Singleton
            s3 = S3Singleton(**s3_session_params).s3

        s3_bucket_name, s3_path = source_dir.split("/", 1)

        s3.meta.client.download_file(s3_bucket_name,
                                     s3_path + "/" + filename,
                                     target_dir + "/" + filename)

        if delete_source:
            s3.delete_objects(Bucket=s3_bucket_name, Delete={'Objects': [s3_path + "/" + filename]})




def load_h2o_model(local_dir, filename, extension=""):
    """
    Loads a saved H2O Model

    :param string local_dir: Local directory where the model is saved
    :param string filename: Filename with which the model is saved
    :param string extension: Extension to the filename with which the model is saved
    :return:
    """

    from h2o import load_model
    return load_model(local_dir + "/" + filename + extension)

def load_mojo_model(local_dir, filename, extension=""):
    """
    Loads a saved H2OMOJOModel (can be used with Spark without a running H2OSparkling Session)

    :param string local_dir: Local directory where the model is saved
    :param string filename: Filename with which the model is saved
    :param string extension: Extension to the filename with which the model is saved
    :return:
    """

    from pysparkling.ml import H2OMOJOModel
    return H2OMOJOModel.create_from_mojo(local_dir + "/" + filename + extension)




"""
Functions that are typically injected with one or more providers and performs a cohesive set of work that might involve
multiple steps
"""

import os
from mercury_ml.common.providers.data_bunch import DataBunch

# source data reading
def read_train_valid_test_data_bunch(read_data_set, train_params, valid_params, test_params):
    """
    Takes the provider "read_data_set" and uses it to read "train, "valid" and "test" DataSets and add them to a
    new DataBunch

    :param callable read_data_set: A function that reads source data and returns a DatSet
    :param dict train_params: The parameters according to which to read the "train" DataSet
    :param dict valid_params: The parameters according to which to read the "valid" DataSet
    :param dict test_params: The parameters according to which to read the "test" DataSet
    :return: a DataBunch
    """
    return DataBunch(
        {
            "train": read_data_set(**train_params),
            "valid": read_data_set(**valid_params),
            "test": read_data_set(**test_params)
        }
    )

def read_test_data_bunch(read_data_set, test_params):
    """
    Takes the provider "read_data_set" and uses it to read a "test" DataSet and add it to a new DataBunch

    :param callable read_data_set: A function that reads source data and returns a DatSet
    :param dict test_params: The parameters according to which to read the "test" DataSet
    :return: a DataBunch
    """
    return DataBunch(
        {
            "test": read_data_set(**test_params)
        }
    )


# artifact storage
def store_artifacts(store_artifact_locally, copy_from_local_to_remote, data, local_dir,
                    filename, remote_dir=None, overwrite_remote=True, keep_local=True):

    """
    Uses "store_artifact_locally" to store artifacts to local disk. Then uses "copy_from_local_to_remote" to also store
    the artifacts in a remote location.

    :param callable store_artifact_locally: A function that stores artifacts to local disk.
    :param callable copy_from_local_to_remote: A function that copies artifacts from local disk to a remote location.
    :param data: The artifact to be stored
    :param string local_dir: The local directory where the artifact should be stored
    :param string filename: The filename under which the artifact should be stored
    :param string remote_dir: The remote "directory" where the artifact should be copied. If None then copying to remote is skipped
    :param bool overwrite_remote: If true then if the artifact already exists in the remote location it will be overwritten
    :param bool keep_local: If false then the local artifact will be deleted after copying to remote
    :return:
    """

    # save to local artifact store
    filepath = store_artifact_locally(data=data,
                                      directory=local_dir,
                                      filename=filename)

    # copy to remote artifact store
    if remote_dir:
        copy_from_local_to_remote(source_dir=local_dir,
                                  target_dir=remote_dir,
                                  filename=os.path.basename(filepath),
                                  overwrite=overwrite_remote,
                                  delete_source=not keep_local)

# metric evaluation
def evaluate_metrics(data_set, custom_metrics_dict):
    """
    Loops through the entries in custom_metrics_dict and calculates each metric on the given dataset

    :param DataSet data_set: A DataSet that has DataWrappers called "targets" and "predictions" as attributes.
    :param dict custom_metrics_dict: A dictionary of custom metrics of the form { {custom_metric_name: custom_metric_function}...}
    where custom_metric_function is a callable that takes the inputs "y_true" and "y_pred" and returns the scalar value of
    the calculated metric.

    :return: A dictionary of the form {{custom_metric_name: metric_value}, ...}
    """
    y_true = data_set.targets.underlying
    y_pred = data_set.predictions.underlying

    custom_metrics = {}

    if not custom_metrics_dict:
        return custom_metrics
    else:
        for custom_metric_name, evaluate_custom_metric in custom_metrics_dict.items():
            custom_metric_value = evaluate_custom_metric(y_true, y_pred)
            custom_metrics[custom_metric_name] = custom_metric_value

        return custom_metrics


def evaluate_label_metrics(data_set, label_specific_custom_metrics_dict):
    """
    Loops through the entries in label_specific_custom_metrics_dict and calculates each metric for each label on the given dataset

    :param DataSet data_set: A DataSet that has DataWrappers called "targets" and "predictions" as attributes.
    :param dict label_specific_custom_metrics_dict: A dictionary of custom label metrics of the form {{custom_label_metric_name: custom_label_metric_function}...}
    where custom_label_metric_function is a callable that takes the inputs "y_true", "y_pred" and "label" and returns a dictionary
    of the form {{label_metric_name: {label_name, label_metric_value}},...}

    :return: A dictionary of the form {{label_metric_name: {label_name, label_metric_value}},...}
    """


    y_true = data_set.targets.underlying
    y_pred = data_set.predictions.underlying
    labels = data_set.targets.field_names

    all_metrics_dict = {}

    if not label_specific_custom_metrics_dict:
        return all_metrics_dict

    for custom_label_metric_name, evaluate_custom_label_metric in label_specific_custom_metrics_dict.items():
        custom_label_metric_dict = evaluate_custom_label_metric(y_true, y_pred, labels)
        all_metrics_dict.update(custom_label_metric_dict)

    return all_metrics_dict



# model storages
def store_model(save_model, model, filename, local_dir, extension=None, copy_from_local_to_remote=None,
                remote_dir=None, overwrite_remote=None, keep_local=True, **kwargs):
    """
    Uses "save_model" and "model" to save a model to local disk. Then uses "copy_from_local_to_remote" to also store
    the model in a remote location.

    :param callable save_model: A function that stores model to local disk.
    :param model: A model object.
    :param string filename: The filename under which the model should be stored
    :param string local_dir: The local directory where the model should be stored
    :param string extension: The extension to the filename (e.g. ".h5", ".pb") with which the model should be saved
    :param callable copy_from_local_to_remote: A function that copies artifacts from local disk to a remote location.
    :param string remote_dir: The remote "directory" where the model should be copied. If None then copying to remote is skipped
    :param bool overwrite_remote: If true then if the model already exists in the remote location it will be overwritten
    :param bool keep_local: If false then the local model will be deleted after copying to remote
    :param kwargs: Additional arguments passed to "save_model"
    :return:
    """

    path = save_model(model=model,
                      filename=filename,
                      local_dir=local_dir,
                      extension=extension,
                      **kwargs)

    if copy_from_local_to_remote and remote_dir:
        copy_from_local_to_remote(source_dir=local_dir,
                                  target_dir=remote_dir,
                                  filename=os.path.basename(path),
                                  overwrite=overwrite_remote,
                                  delete_source=not keep_local)

    return path

def load_model(load_model, filename, local_dir, extension=None, remote_dir=None, copy_from_remote_to_local=None,
               always_fetch_remote=False, **kwargs):
    """
    First checks if the desired model is avaible at the local location. If not it is copied from remote to local. Thereafter
    the model is loaded and returned.

    :param callable load_model: Function that will return a loaded model
    :param string filename: The filename under which the model is stored
    :param string local_dir: The local directory where the model is stored
    :param string extension: The extension to the filename (e.g. ".h5", ".pb") with which the model is saved
    :param string remote_dir: The remote "directory" where the model is saved.
    :param callable copy_from_remote_to_local: Function that copies data from a remote location to a local one
    :param bool always_fetch_remote: If true the the model will always be fetched from the remote location, even if
    it appears to be present locally
    :param kwargs: Additional arguments that are passed to "load_model"
    :return: A model object
    """

    path = os.path.join(local_dir, filename + extension)

    if not os.path.isfile(path) or always_fetch_remote:
        copy_from_remote_to_local(source_dir=remote_dir,
                                  target_dir=local_dir,
                                  filename=filename + extension,
                                  overwrite=True,
                                  delete_source=False)

    model = load_model(filename=filename,
                       local_dir=local_dir,
                       extension=extension,
                       **kwargs)

    return model



"""ImageGenerator and DirectoryIterator classes used for models with multiple (attribute-type) labels"""

from keras.preprocessing.image import ImageDataGenerator, DirectoryIterator
import os
import numpy as np
import pandas as pd
from keras import backend as K
from keras.preprocessing.image import load_img, img_to_array, array_to_img


class MultiLabelImageDataGenerator(ImageDataGenerator):
    """
    An image data generator that is able to return multiple labels, based on a LabelMetaStore
    """
    def __init__(self, label_df, *args, **kwargs):
        self.label_df = label_df
        super(MultiLabelImageDataGenerator, self).__init__(*args, **kwargs)

    def flow_from_directory(self, *args, **kwargs):
        kwargs["image_data_generator"] = self
        kwargs["label_df"] = self.label_df
        return MultiLabelDirectoryInterator(*args, **kwargs)


class MultiLabelDirectoryInterator(DirectoryIterator):
    def __init__(self, *args, **kwargs):
        self.label_df = kwargs.pop("label_df")
        super(MultiLabelDirectoryInterator, self).__init__(*args, **kwargs)

    def get_target_dummies(self):
        """
        :return: array, the targets associated with this generator, in one-hot-encoded (dummy) format
        """

        if self.shuffle:
            raise AssertionError(
                "The 'shuffle' parameter should always be switched to 'False' for a generator for which the array of "
                "targets is retrieved!")
        else:
            filenames = [self._split_filenames_vector(filename) for filename in self.filenames]
            targets_df = self.label_df.loc[filenames]
            return targets_df.values

    def get_labels_dummies(self):
        return list(self.label_df.columns)

    def get_labels_classes(self):
        return list(self.label_df.columns)


    def get_target_classes(self):
        """
        :return: array, the targets associated with this generator, in dense format
        """
        # for multi-label models there is no single "class", therefore an array of "-1" values is returned
        return np.full(len(self.classes), -1)

    def get_filenames(self):
        """
        Gets the unique filenames associated with each of the source images in a generator

        :param generator: DirectoryIterator, a generator used to feed the model with training images
        :return: An array of strings
        """

        if self.shuffle:
            raise AssertionError( """The 'shuffle' parameter should always be switched to 'False' for a generator for 
            which the array of filenames is retrieved!""")

        return self.filenames

    def _split_filenames_vector(self, filename):
        _, fname_without_folder = os.path.split(filename)
        fname_name_only, _ = os.path.splitext(fname_without_folder)
        return fname_name_only

    def _load_label_array(self, fname_name_only):
        return self.label_df.loc[fname_name_only].values

    def _get_batches_of_transformed_samples(self, index_array):
        batch_x = np.zeros((len(index_array),) +
                           self.image_shape, dtype=K.floatx())
        grayscale = self.color_mode == "grayscale"

        # custom parameter
        batch_y = np.zeros(
            (len(index_array), self.label_df.shape[1]), dtype=K.floatx())

        # build batch of image data
        for i, j in enumerate(index_array):
            fname = self.filenames[j]
            img = load_img(os.path.join(self.directory, fname),
                           grayscale=grayscale,
                           target_size=self.target_size,
                           interpolation=self.interpolation)
            x = img_to_array(img, data_format=self.data_format)
            x = self.image_data_generator.random_transform(x)
            x = self.image_data_generator.standardize(x)

            # custom fetch
            path, fname_without_folder = os.path.split(fname)
            fname_name_only, extension = os.path.splitext(fname_without_folder)
            y = self._load_label_array(fname_name_only)

            batch_x[i] = x
            batch_y[i] = y

        # optionally save augmented images to disk for debugging purposes
        if self.save_to_dir:
            for i, j in enumerate(index_array):
                img = array_to_img(batch_x[i], self.data_format, scale=True)
                fname = "{prefix}_{index}_{hash}.{format}".format(prefix=self.save_prefix,
                                                                  index=j,
                                                                  hash=np.random.randint(1e7),
                                                                  format=self.save_format)
                img.save(os.path.join(self.save_to_dir, fname))

        return batch_x, batch_y



"""ImageGenerator and DirectoryIterator classes used for models with single inputs"""

from keras.preprocessing.image import ImageDataGenerator, DirectoryIterator
import pandas as pd


class SingleInputImageDataGenerator(ImageDataGenerator):
    """
    Wraps the vanilla Keras ImageDataGenerator into a slightly altered class definition
    """
    def __init__(self, *args, **kwargs):
        super(SingleInputImageDataGenerator, self).__init__(*args, **kwargs)

    def flow_from_directory(self, *args, **kwargs):
        kwargs["image_data_generator"] = self
        return SingleInputDirectoryInterator(*args, **kwargs)


class SingleInputDirectoryInterator(DirectoryIterator):
    def __init__(self, *args, **kwargs):
        super(SingleInputDirectoryInterator, self).__init__(*args, **kwargs)

    def get_labels_dummies(self):
        classes_lookup = {v: k for k, v in self.class_indices.items()}
        return [v for k,v in sorted(classes_lookup.items())] #returns a list, sorted on the class number. #TODO test this

    def get_labels_classes(self):
        return ["class"]

    def get_target_dummies(self):
        """
        :return: array, the targets associated with this generator, in one-hot-encoded (dummy) format
        """
        if self.shuffle:
            raise AssertionError(
                "The 'shuffle' parameter should always be switched to 'False' for a generator for which the array of targets is retrieved!")
        else:
            df = pd.get_dummies(pd.Series(self.classes)).to_dense()
            return df.values

    def get_target_classes(self):
        """
        :return: array, the targets associated with this generator, in dense format
        """
        if self.shuffle:
            raise AssertionError(
                "The 'shuffle' parameter should always be switched to 'False' for a generator for which the array of targets is retrieved!")

        else:
            return self.classes

    def get_filenames(self):
        """
        Gets the unique filenames associated with each of the source images in a generator

        :param generator: DirectoryIterator, a generator used to feed the model with training images
        :return: An array of strings
        """

        if self.shuffle:
            raise AssertionError("""The 'shuffle' parameter should always be switched to 'False' for a generator for which the array of filenames is retrieved!""")

        return self.filenames





"""
Simple IoC containers that provide direct access to various providers
"""


class SourceReaders:
    """IoC container for SourceStore providers."""

    from mercury_ml.common.providers.source_reading import disk, hive

    read_disk_pandas = disk.read_pandas_data_set
    read_disk_keras_single_input_iterator = disk.read_keras_single_input_image_iterator_data_set
    read_disk_keras_multi_label_iterator = disk.read_keras_multi_label_image_iterator_data_set
    read_hive_spark = hive.read_spark_data_set

class CustomMetrics:
    """IoC container for CustomMetric providers."""

    from mercury_ml.common.providers.metric_evaluation.numpy import metrics as numpy_metrics

    evaluate_numpy_auc = numpy_metrics.evaluate_macro_auc
    evaluate_numpy_micro_auc = numpy_metrics.evaluate_micro_auc
    evaluate_numpy_macro_auc = numpy_metrics.evaluate_macro_auc


class CustomLabelMetrics:
    """IoC container for CustomLabelMetric providers."""
    
    from mercury_ml.common.providers.metric_evaluation.numpy import label_metrics as numpy_label_metrics

    evaluate_numpy_auc = numpy_label_metrics.evaluate_auc
    evaluate_numpy_confusion_matrix = numpy_label_metrics.evaluate_confusion_matrix
    evaluate_numpy_accuracy = numpy_label_metrics.evaluate_accuracy
    evaluate_numpy_multi_label_confusion_matrix = numpy_label_metrics.evaluate_multi_label_confusion_matrix


class LocalArtifactStorers:
    from mercury_ml.common.providers.artifact_storage import local

    store_pandas_pickle = local.store_pandas_pickle
    store_pandas_json = local.store_pandas_json
    store_dict_json = local.store_dict_json
    store_h2o_frame = local.store_h2o_frame

class ArtifactCopiers:
    """IoC container of FileMover providers for moving saved models from local to remote store"""

    from mercury_ml.common.providers.artifact_copying import from_disk, from_gcs, from_s3, from_hdfs

    copy_from_disk_to_disk = from_disk.copy_from_disk_to_disk

    copy_from_disk_to_hdfs = from_disk.copy_from_disk_to_hdfs
    copy_from_disk_to_s3 = from_disk.copy_from_disk_to_s3
    copy_from_disk_to_gcs = from_disk.copy_from_disk_to_gcs

    copy_from_s3_to_disk = from_s3.copy_from_s3_to_disk
    copy_from_hdfs_to_disk = from_hdfs.copy_from_hdfs_to_disk
    copy_from_gcs_to_disk = from_gcs.copy_from_gcs_to_disk




class SparkDataWrapper():

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names

    def slice_on_column_names(self, column_names):
        """
        Returns a Spark DataFrame consisting of the columns that are present in the column_names list

        :param list list column_names: the names of the columns to return
        :return: Spark DataFrame consisting of the selected columns
        """
        return self.underlying[column_names]

    def to_h2o_sparkling(self, factor_columns_list=None, h2o_sparkling_params=None):
        """
        Creates a new H2OSparklingDataWrapper based on this SparkDataWrapper and returns it as a new object

        :param list factor_columns_list: The names of the columns that should be treated as factors
        :param list h2o_sparkling_params: A dictionary of parameters according to which to get or initialize an H2OSparkling session

        :return: a new H2OSparklingDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.h2o import H2OSparklingDataWrapper
        from mercury_ml.h2o.providers.session import get_or_create_h2o_sparkling

        # get H2OSparkling context (#TODO: if context doesn't exist yet, raise error! Should not be created here)
        if not h2o_sparkling_params:
            h2o_sparkling_params = {}
        h2o_context = get_or_create_h2o_sparkling(**h2o_sparkling_params)

        # convert to H2OFrame
        df_h2o = h2o_context.as_h2o_frame(self.underlying)

        # convert factor columns
        if not factor_columns_list:
            factor_columns_list = []

        for factor_column_name in factor_columns_list:
            df_h2o[factor_column_name] = df_h2o[factor_column_name].asfactor()

        return H2OSparklingDataWrapper(df_h2o, self.field_names)

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this SparkDataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        return self.to_pandas().to_numpy()


    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this SparkDataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        df = self.underlying.toPandas()
        return PandasDataWrapper(df, self.field_names)

    def to_spark(self):
        return self

    def concatenate(self, right_data_wrapper):
        """
        Concatenates this SparkDataWrapper with the one given in right_data_wrapper and returns a new SparkDataWrapper

        :param SparkDataWrapper right_data_wrapper: to be concatenated to the right of "self"
        :return: a new SparkDataWrapper
        """

        raise NotImplementedError("Concatenating Spark DFs not implemented yet")

        # spark = _get_spark()
        #
        # new_underlying = ... # TODO how to concatenate while maintaining row order?
        # new_field_names = self.field_names + right_data_wrapper.field_names
        # return SparkDataWrapper(
        #     underlying = new_underlying,
        #     field_names = new_field_names
        # )


def fit(model, data_bunch, **kwargs):

    """
    Fits an H2O Model
    :param model: An H2O Model
    :param data_bunch: A DataBunch with "train" and "valid" datasets that consist of H2ODataWrapper or H2OSparklingDataWrapper
    :param kwargs: Keyword arguments to be passed to model.train
    :return: A (fitted) H2O Model
    """

    train_df = data_bunch.train.full_data.underlying
    valid_df = data_bunch.valid.full_data.underlying
    features_list = data_bunch.train.features.field_names
    targets_list = data_bunch.valid.targets.field_names[0]

    model.train(x=features_list,
                y=targets_list,
                training_frame=train_df,
                validation_frame=valid_df,
                **kwargs
                )

    return model





class H2ODataWrapper():
    """
    A DataWrapper with an H2OFrame as its underlying data
    """

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names

    def slice_on_column_names(self, column_names):
        """
        Returns an H2OFrame consisting of the columns that are present in the column_names list

        :param list column_names: the names of the columns to return
        :return: H2OFrame consisting of the selected columns
        """

        return self.underlying[column_names]

    def to_pandas(self):
        """
        Creates a new PandasDataWrapper based on this H2ODataWrapper and returns it as a new object

        :return: a new PandasDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.pandas import PandasDataWrapper
        return PandasDataWrapper(self.underlying.as_data_frame(use_pandas=True), self.field_names)

    def to_numpy(self):
        """
        Creates a new NumpyDataWrapper based on this H2ODataWrapper and returns it as a new object

        :return: a new NumpyDataWrapper
        """

        from mercury_ml.common.providers.data_wrappers.numpy import NumpyDataWrapper
        return NumpyDataWrapper(self.underlying.as_data_frame(use_pandas=True).values, self.field_names)

    def to_h2o(self):
        return self

    def concatenate(self, right_data_wrapper):
        """
        Concatenates this H2ODataWrapper with the one given in right_data_wrapper and returns a new H2ODataWrapper

        :param H2ODataWrapper right_data_wrapper: to be concatenated to the right of "self"
        :return: a new H2ODataWrapper
        """

        new_underlying =self.underlying.cbind(right_data_wrapper.underlying)
        new_field_names = self.field_names + right_data_wrapper.field_names
        return H2ODataWrapper(
            underlying = new_underlying,
            field_names = new_field_names
        )


class H2OSparklingDataWrapper():
    """
    A DataWrapper with an H2OFrame within an H2OSparkling session) as its underlying data
    """

    def __init__(self, underlying, field_names):
        self.underlying = underlying
        self.field_names = field_names

    def slice_on_column_names(self, column_names):
        """
        Returns an H2OFrame consisting of the columns that are present in the column_names list

        :param list column_names: the names of the columns to return
        :return: H2OFrame consisting of the selected columns
        """

        return self.underlying[column_names]

    def to_pandas(self, h2o_sparkling_params=None):
        """
        Creates a new PandasDataWrapper based on this H2OSparklingDataWrapper and returns it as a new object

        :param: dict h2o_sparkling_params: Parameters passed to "get_or_create_h2o_sparkling" which will get or initiate Spark and H2OSparkling sessions
        :return: a new PandasDataWrapper
        """

        return self.to_spark(h2o_sparkling_params).to_pandas()

    def to_numpy(self, h2o_sparkling_params=None):
        """
        Creates a new NumpyDataWrapper based on this H2OSparklingDataWrapper and returns it as a new object

        :param: dict h2o_sparkling_params: Parameters passed to "get_or_create_h2o_sparkling" which will get or initiate Spark and H2OSparkling sessions
        :return: a new NumpyDataWrapper
        """

        return self.to_spark(h2o_sparkling_params).to_pandas().to_numpy()

    def to_spark(self, h2o_sparkling_params=None):
        """
        Creates a new SparkDataWrapper based on this H2OSparklingDataWrapper and returns it as a new object

        :param: dict h2o_sparkling_params: A dictionary of parameters according to which to get or initialize an H2OSparkling session
        :return: a new SparkDataWrapper
        """

        from mercury_ml.h2o.providers.session import get_or_create_h2o_sparkling
        from mercury_ml.common.providers.data_wrappers.spark import SparkDataWrapper

        # get H2OSparkling context #TODO: if context doesn't exist yet, raise error! Should not be created here
        if not h2o_sparkling_params:
            h2o_sparkling_params = {}
        h2o_context = get_or_create_h2o_sparkling(**h2o_sparkling_params)

        return SparkDataWrapper(
            underlying=h2o_context.as_spark_frame(self.underlying),
            field_names=self.field_names
        )

    def concatenate(self, right_data_wrapper):
        """
        Concatenates this H2OSparklingDataWrapper with the one given in right_data_wrapper and returns a new H2ODataWrapper

        :param H2OSparklingDataWrapper right_data_wrapper: to be concatenated to the right of "self"
        :return: a new H2OSparklingDataWrapper
        """

        new_underlying=self.underlying.cbind(right_data_wrapper.underlying)
        new_field_names=self.field_names + right_data_wrapper.field_names
        return H2OSparklingDataWrapper(
            underlying=new_underlying,
            field_names=new_field_names
        )
